#include <stdio.h>
#include <stdlib.h>
#include "opdef.h"

typedef unsigned long u_long;

static char *regname[32] = {
	"zero","at","v0","v1","a0","a1","a2","a3",
	"t0","t1","t2","t3","t4","t5","t6","t7",
	"s0","s1","s2","s3","s4","s5","s6","s7",
	"t8","t9","k0","k1","gp","sp","fp","ra"
};

static char *cop0regname[32] = {
	/* from mips 4400 manual
	"Index","Random","EntryLo0","EntryLo1","Context","PageMask","Wired","$7",
	"BadVAddr","Count","EntryHi","Compare","Status","Cause","EPC","PRId",
	"Config","LLAddr","WatchLo","WatchHi","XContext","$21","$22","$23",
	"$24","$25","ECC","CacheErr","TagLo","TagHi","ErrorEPC","$31"
	*/
	"INDEX","RAND","$2","BPC","$3","BDA","TAR","DCIC",
	"BadVAddr","BDAM","$10","BPCM","SR","CAUSE","EPC","PRId",
	"$16","$17","$18","$19","$20","$21","$22","$23",
	"$24","$25","$26","$27","$28","$29","$30","$31"
};

static char *cop2cregname[32] = {
	"R11R12","R13R21","R22R23","R31R32","R33","TRX","TRY","TRZ",
	"L11L12","L13L21","L22L23","L31L32","L33","RBK","GBK","BBK",
	"LR1LR2","LR3LG1","LG2LG3","LB1LB2","LB3","RFC","GFC","BFC",
	"OFX","OFY","H","DQA","DQB","ZSF3","ZSF4","FLAG"
};

static char *cop2regname[32] = {
	"VXY0","VZ0","VXY1","VZ1","VXY2","VZ2","RGB","OTZ",
	"IR0","IR1","IR2","IR3","SXY0","SXY1","SXY2","SXYP",
	"SZ0","SZ1","SZ2","SZ3","RGB0","RGB1","RGB2","$23",
	"MAC0","MAC1","MAC2","MAC3","IRGB","ORGB","LZCS","LZCR"
};

static char *cregname[32] = {
	"$0","$1","$2","$3","$4","$5","$6","$7",
	"$8","$9","$10","$11","$12","$13","$14","$15",
	"$16","$17","$18","$19","$20","$21","$22","$23",
	"$24","$25","$26","$27","$28","$29","$30","$31"
};

static char *nemonic[64] = {
	"special","bcond","j","jal","beq","bne","blez","bgtz",
	"addi","addiu","slti","sltiu","andi","ori","xori","lui",
	"cop0","cop1","cop2","cop3","","","","",
	"","","","","","","","",
	"lb","lh","lwl","lw","lbu","lhu","lwr","",
	"sb","sh","swl","sw","","","swr","",
	"lwc0","lwc1","lwc2","lwc3","","","","",
	"swc0","swc1","swc2","swc3","","","",""
};

static char *special[64] = {
	"sll","","srl","sra","sllv","","srlv","srav",
	"jr","jalr","","","syscall","break","","",
	"mfhi","mthi","mflo","mtlo","","","","",
	"mult","multu","div","divu","","","","",
	"add","addu","sub","subu","and","or","xor","nor",
	"","","slt","sltu","","","","",
	"","","","","","","","",
	"","","","","","","",""
};

static char *bcond[32] = {
	"bltz","bgez","","","","","","",
	"","","","","","","","",
	"bltzal","bgezal","","","","","","",
	"","","","","","","",""
};

static char *cop[16] = {
	"mfc","","cfc","","mtc","","ctc","",
	"bc","","","","bc","","",""
};

static char *cop0[32] = {
	"","tlbr","tlbwi","","","","tlbwr","",
	"tlbp","","","","","","","",
	"rfe","","","","","","","",
	"","","","","","","",""
};

static char cop0reg[32] = {
	"INDEX","RAND","","","","","","",
	"BadVAddr","","","","SR","CAUSE","EPC","PRId",
	"","","","","","","","",
	"","","","","","","","",
};

#define	MASK(from,n)	(((1<<(n))-1)<<(from))
#define	CHECK(a)

void disasm(char *buf,u_long code,u_long addr)
{
	int	op,rs,rt,rd,immS,func;
	if (!code) {
		strcpy(buf,"nop");
		return;
	}

	op = code>>26;
	rs = (code>>21)&31;
	rt = (code>>16)&31;
	rd = (code>>11)&31;
	immS = (short)code;

	switch(op){
	case SPECIAL:
		func = code&63;
		switch(func){
		case SLL:
		case SRL:
		case SRA:
			CHECK(rs) /* err */
			sprintf(buf,"%-5s %s,%s,%d",special[func],regname[rd],regname[rt],(code>>6)&31);
			break;
		case SLLV:
		case SRLV:
		case SRAV:
			CHECK(code & MASK(6,5)) /* err */
			sprintf(buf,"%-5s %s,%s,%s",special[func],regname[rd],regname[rt],regname[rs]);
			break;
		case JR:
			CHECK(code & MASK(6,15)) /* err */
			sprintf(buf,"%-5s %s",special[func],regname[rs]);
			break;
		case JALR:
			CHECK(code & (MASK(6,5)|MASK(16,5))) /* err */
			sprintf(buf,"%-5s %s,%s",special[func],regname[rs],regname[rd]);
			break;
		case SYSCALL:
		case BREAK:
			sprintf(buf,"%-5s %08x",special[func],(code>>6)&0xfffff);
			break;
		case MFHI:
		case MFLO:
			CHECK(code & (MASK(6,5)|MASK(16,10))) /* err */
			sprintf(buf,"%-5s %s",special[func],regname[rd]);
			break;
		case MTHI:
		case MTLO:
			CHECK(code & MASK(6,15)) /* err */
			sprintf(buf,"%-5s %s",special[func],regname[rs]);
			break;
		case MULT:
		case MULTU:
		case DIV:
		case DIVU:
			CHECK(code & MASK(6,10)) /* error */
			sprintf(buf,"%-5s %s,%s",special[func],regname[rs],regname[rt]);
			break;
		case ADD:
		case ADDU:
		case SUB:
		case SUBU:
		case AND:
		case OR:
		case XOR:
		case NOR:
		case SLT:
		case SLTU:
			CHECK(code & MASK(6,5)) /* err */
			sprintf(buf,"%-5s %s,%s,%s",special[func],regname[rd],regname[rs],regname[rt]);
			break;
		default:
			strcpy(buf,"unknown special");
			break;
		}
		break;
	case BCOND:
		sprintf(buf,"%-5s %s,%08x",bcond[rt],regname[rs],addr+(immS+1)*4);
		break;
	case J:
	case JAL:
		sprintf(buf,"%-5s %08x",nemonic[op],(addr&0xf0000000)|(code&0x3ffffff)*4);
		break;
	case BEQ:
	case BNE:
		sprintf(buf,"%-5s %s,%s,%08x",nemonic[op],regname[rs],regname[rt],addr+(immS+1)*4);
		break;
	case BLEZ:
	case BGTZ:
		CHECK(rt) /* err */
		sprintf(buf,"%-5s %s,%08x",nemonic[op],regname[rs],addr+(immS+1)*4);
		break;
	case ADDI:
	case ADDIU:
	case SLTI:
	case SLTIU:
	case ANDI:
	case ORI:
	case XORI:
	case LUI:
		sprintf(buf,"%-5s %s,%s,%d\t# %04x",nemonic[op],regname[rt],regname[rs],immS,code&0xffff);
		break;
	case COP0:
		if (rs&0x10) {
			CHECK(code && MASK(5,20)) /* err */
			sprintf(buf,"%s",cop0[code&31]);
		} else {
			CHECK(code & MASK(0,11)) /* err */
			switch(rs) {
			case MFC:
			case MTC:
				sprintf(buf,"%s0  %s,%s",cop[rs],regname[rt],cop0regname[rd]);
				break;
			case CFC:
			case CTC:
				sprintf(buf,"%s0  %s,%s",cop[rs],regname[rt],cregname[rd]);
				break;
			}
		}
		break;
	case COP2:
		if (rs&0x10) {
			sprintf(buf,"cop%d  %08x",op&3,code&0x1ffffff);
		} else {
			CHECK(code & MASK(0,11)) /* err */
			switch(rs) {
			case MFC:
			case MTC:
				sprintf(buf,"%s2  %s,%s",cop[rs],regname[rt],cop2regname[rd]);
				break;
			case CFC:
			case CTC:
				sprintf(buf,"%s2  %s,%s",cop[rs],regname[rt],cop2cregname[rd]);
				break;
			}
		}
		break;
	case LB:
	case LH:
	case LWL:
	case LW:
	case LBU:
	case LHU:
	case LWR:
	case SB:
	case SH:
	case SWL:
	case SW:
	case SWR:
		sprintf(buf,"%-5s %s,%d(%s)",nemonic[op],regname[rt],immS,regname[rs]);
		break;
	case LWC0:
	case SWC0:
		sprintf(buf,"%-5s %s,%d(%s)",nemonic[op],cop0regname[rt],immS,regname[rs]);
		break;
	case SWC2:
	case LWC2:
		sprintf(buf,"%-5s %s,%d(%s)",nemonic[op],cop2regname[rt],immS,regname[rs]);
		break;
	case LWC1:
	case LWC3:
	case SWC1:
	case SWC3:
		sprintf(buf,"%-5s %s,%d(%s)",nemonic[op],cregname[rt],immS,regname[rs]);
		break;
	default:
		strcpy(buf,"unknown");
		break;
	}
}

#ifdef MAKEDIS
typedef struct _EXE_HEADER_ {
	char id[8];
	u_long text;			/* SCE only */
	u_long data;			/* SCE only */
	u_long pc0;
	u_long gp0;			/* SCE only */
	u_long t_addr;
	u_long t_size;
	u_long d_addr;			/* SCE only */
	u_long d_size;			/* SCE only */
	u_long b_addr;			/* SCE only */
	u_long b_size;			/* SCE only */
	u_long s_addr;
	u_long s_size;
	u_long SavedSP;
	u_long SavedFP;
	u_long SavedGP;
	u_long SavedRA;
	u_long SavedS0;
} EXE_HEADER;

int main(int argc,char *argv[])
{
	FILE *fp;
	char *buf,*mem;
	u_long addr,startaddr,fsize;
	int	offset,i;
	char inputbuf[256];

	EXE_HEADER *head;
	char strbuf[256];

	fp = fopen(argv[1],"rb");
	fseek(fp,0,SEEK_END);
	fsize = ftell(fp);
	fseek(fp,0,SEEK_SET);
	buf = malloc(fsize);
	fread(buf,1,fsize,fp);
	fclose(fp);

	head = (void*)buf;
	mem = buf;
	if (memcmp(head->id,"PS-X EXE",8)==0) {
		startaddr = head->t_addr;
		mem+=2048;
	} else {
		startaddr = 0xbfc00000;
	}

	mem-=startaddr;
	addr = startaddr;


   while(1) {
	for(i=0;i<16;i++) {
		u_long code = *(u_long*)(mem+addr);
		disasm(strbuf,code,addr);
		printf("%08x %08x %s\n",addr,code,strbuf);
		addr+=4;
	}
	printf(">");
	gets(inputbuf);
	if (strcmp(inputbuf,"q")==0) break;
    switch(inputbuf[0]){
    case 'a': { char *endp; addr=strtoul(inputbuf+1,&endp,16); }
    }
   }
}
#endif
