#include <stdio.h>
#include <stdlib.h>

#ifndef TRUE
#define	TRUE	1
#define	FALSE	0
#endif

unsigned char *rom,*ram,*scratchpad,*extrom;
unsigned short *vram,*spumem;
extern int stop,disasmflg;
int emulate_bios;
int verbose;

void memalloc(void)
{
	rom = malloc(0x80000);
	extrom = malloc(0x20000);
	memset(extrom,0xff,0x20000);
	ram = malloc(0x200000);
	scratchpad = malloc(0x1000);
//	vram = malloc(1024*512*2);
	spumem = malloc(512*1024);
}

void memfree(void)
{
	free(extrom);
	free(rom);
	free(ram);
	free(scratchpad);
//	free(vram);
	free(spumem);
}

void go(void)
{
	while(!stop) {
		if (update_counter()) Interrupt();
		doInst();
	}
}

typedef struct EXEC {
	char id[8];
	unsigned long text;			/* SCE only */
	unsigned long data;			/* SCE only */
	unsigned long pc0;
	unsigned long gp0;			/* SCE only */
	unsigned long t_addr;
	unsigned long t_size;
	unsigned long d_addr;			/* SCE only */
	unsigned long d_size;			/* SCE only */
	unsigned long b_addr;			/* SCE only */
	unsigned long b_size;			/* SCE only */
	unsigned long s_addr;
	unsigned long s_size;
	unsigned long SavedSP;
	unsigned long SavedFP;
	unsigned long SavedGP;
	unsigned long SavedRA;
	unsigned long SavedS0;
	char pad[2048-19*4];
} EXE_HEADER;

extern int reg[];

static char helpmsg[] = 
	"fpse [-option] [file]\n"
	" file         PSX-EXE file\n"
	" -c           compiler(default)\n"
	" -i           interpriter\n"
	" -b<biosfile> bios filename\n"
	" -v           verbose\n"
	" -d           disassemble\n"
	"run PSX-EXE with clone bios, or original bios from start\n"
	"to stop, press ctrl+C\n";

int main(int argc,char *argv[])
{
	FILE *fp;
	char *file = NULL;
	char *biosfile = NULL;
	int i;
	int compile = TRUE;
	disasmflg = FALSE;
	verbose = FALSE;

	printf("PlayStation Emulator by BERO\n");
	if (argc<2) {
		printf(helpmsg);
		return -1;
	}

	for(i=1;i<argc;i++) {
		char *s = argv[i];
		if (s[0]=='-') {
			switch(s[1]) {
			case 'c': compile = TRUE; break;
			case 'i': compile = FALSE; break;
			case 'v': verbose = TRUE; break;
			case 'd': disasmflg = TRUE; break;
			case 'b': biosfile = s+2; break;
			default:
				printf(helpmsg);
				return -1;
			}
		} else {
			file = s;
		}
	}

	memalloc();
	if (biosfile) {
	fp = fopen(biosfile,"rb");
	if (fp==NULL) {
		printf("can't load biosfile\n");
	} else {
		fread(rom,1,0x80000,fp);
		fclose(fp);
	}
	}

	win_init();
	vram = win_createbitmap(1024,512,16);

	{int i;
	for(i=0;i<1024*512;i++) vram[i]=i;
	}

	compile_init();

	aspi_init();
	hw_init();
	Reset();

	if ((fp = fopen("libps.exe","rb"))!=NULL) {
		struct EXEC head;
		fread(&head,1,sizeof(head),fp);
		if (memcmp(head.id,"PS-X EXE",8)==0) {
			printf("%x-%x\n",head.t_addr,head.t_size);
			fread(&ram[head.t_addr&0x1fffff],1,head.t_size,fp);
		}
		fclose(fp);
	}


	if (file && (fp = fopen(file,"rb"))!=NULL) {
		struct EXEC head;
		fread(&head,1,sizeof(head),fp);
		if (memcmp(head.id,"PS-X EXE",8)==0) {
			fread(&ram[head.t_addr&0x1fffff],1,head.t_size,fp);
			setpc(head.pc0);
			emulate_bios = 1;
			reg[30] = reg[29] = 0x801fff00; /* defailt stack */
		}
		fclose(fp);
	}

	if (compile)
		compile_go();
	else
		go();
	memfree();
}
