#include	<stdio.h>
#include	"type.h"

void hw_write32(UINT32 adr,UINT32 data);
void hw_write16(UINT32 adr,UINT32 data);
void hw_write8(UINT32 adr,UINT32 data);
UINT32 hw_read32(UINT32 adr);
UINT16 hw_read16(UINT32 adr);
UINT8 hw_read8(UINT32 adr);

extern unsigned char *rom,*ram,*scratchpad;
extern int stop;

void printpc(void);

extern int verbose;
static int writeok=1;
static void mem_ctrl(UINT32 data)
{
	/* LbVtbVpH */
	if (verbose) printf("memctrl %08x\n",data);
	switch(data) {
	case 0x00000800:
	case 0x00000804:
		/* LbV 0x0000000? */
		writeok = 0;
		break;
	case 0x0001e988:
		/* C */
		writeok = 1;
		break;
	default:
		printf("unknown memctrl %08x\n",data);
		break;
	}
}

UINT8 read8(UINT32 adr)
{
	if((adr>=0x80000000 && adr<0x80200000)
	|| (adr>=0xa0000000 && adr<0xa0200000)
	|| (adr>=0x00000000 && adr<0x00200000)) return ram[adr&0x1fffff];

	else if (adr>=0xbfc00000 && adr<0xbfc80000) return rom[adr&0x7ffff];
	else if (adr>=0x1f800000 && adr<0x1f801000) return scratchpad[adr&0xfff];
	else if (adr>=0x1f801000 && adr<0x1f808000) return hw_read8(adr);
	else { printpc();printf("can't read %08x\n",adr); return -1; }
}

UINT16 read16(UINT32 adr)
{
	if((adr>=0x80000000 && adr<0x80200000)
	|| (adr>=0xa0000000 && adr<0xa0200000)
	|| (adr>=0x00000000 && adr<0x00200000)) return *(UINT16*)&ram[adr&0x1fffff];

	else if (adr>=0xbfc00000 && adr<0xbfc80000) return *(UINT16*)&rom[adr&0x7ffff];
	else if (adr>=0x1f800000 && adr<0x1f801000) return *(UINT16*)&scratchpad[adr&0xfff];
	else if (adr>=0x1f801000 && adr<0x1f808000) return hw_read16(adr);
	else { printpc();printf("can't read %08x\n",adr); return -1; }
}

UINT32 read32(UINT32 adr)
{
	if((adr>=0x80000000 && adr<0x80200000)
	|| (adr>=0xa0000000 && adr<0xa0200000)
	|| (adr>=0x00000000 && adr<0x00200000)) return *(UINT32*)&ram[adr&0x1fffff];

	else if (adr>=0xbfc00000 && adr<0xbfc80000) return *(UINT32*)&rom[adr&0x7ffff];
	else if (adr>=0x1f800000 && adr<0x1f801000) return *(UINT32*)&scratchpad[adr&0xfff];
	else if (adr>=0x1f801000 && adr<0x1f808000) return hw_read32(adr);
	else { printpc();printf("can't read %08x\n",adr); return -1; }
}

void write8(UINT32 adr,UINT32 data)
{
	if((adr>=0x80000000 && adr<0x80200000)
	|| (adr>=0xa0000000 && adr<0xa0200000)
	|| (adr>=0x00000000 && adr<0x00200000)) ram[adr&0x1fffff]=data;

//	else if (adr>=0xbfc00000 && adr<0xbfc80000) rom[adr&0x7ffff]=data;
	else if (adr>=0x1f800000 && adr<0x1f801000) scratchpad[adr&0xfff]=data;
	else if (adr>=0x1f801000 && adr<0x1f808000) hw_write8(adr,data);
	else { printpc();printf("can't write %08x,%02x\n",adr,data);}
}

void write16(UINT32 adr,UINT32 data)
{
	if((adr>=0x80000000 && adr<0x80200000)
	|| (adr>=0xa0000000 && adr<0xa0200000)
	|| (adr>=0x00000000 && adr<0x00200000)) *(UINT16*)&ram[adr&0x1fffff]=data;

//	else if (adr>=0xbfc00000 && adr<0xbfc80000) *(UINT16*)&rom[adr&0x7ffff]=data;
	else if (adr>=0x1f800000 && adr<0x1f801000) *(UINT16*)&scratchpad[adr&0xfff]=data;
	else if (adr>=0x1f801000 && adr<0x1f808000) hw_write16(adr,data);
	else { printpc();printf("can't write %08x,%04x\n",adr,data);}
}

extern int writebreak;

void write32(UINT32 adr,UINT32 data)
{
	if((adr>=0x80000000 && adr<0x80200000)
	|| (adr>=0xa0000000 && adr<0xa0200000)
	|| (adr>=0x00000000 && adr<0x00200000)) { if (writeok) *(UINT32*)&ram[adr&0x1fffff]=data; }

//	else if (adr>=0xbfc00000 && adr<0xbfc80000) *(UINT32*)&rom[adr&0x7ffff]=data;
	else if (adr>=0x1f800000 && adr<0x1f801000) *(UINT32*)&scratchpad[adr&0xfff]=data;
	else if (adr>=0x1f801000 && adr<0x1f808000) hw_write32(adr,data);
	else if (adr==0xfffe0130) mem_ctrl(data);
	else { printpc();printf("can't write %08x,%08x\n",adr,data);}
/*
	if (adr == writebreak) {
		printpc();printf("write break %08x,%08x\n",adr,data);
	}
*/
}

void *baseaddr(UINT32 adr)
{
	if((adr>=0x80000000 && adr<0x80200000)
	|| (adr>=0xa0000000 && adr<0xa0200000)
	|| (adr>=0x00000000 && adr<0x00200000)) return ram - (adr&~0x1fffff);

	else if (adr>=0xbfc00000 && adr<0xbfc80000) return rom - 0xbfc00000;
	else if (adr>=0x1f800000 && adr<0x1f801000) return scratchpad - 0x1f800000;
	else { stop=1; printpc();printf("can't exec %08x\n",adr);}
}

void *baseaddr2(UINT32 adr)
{
	if((adr>=0x80000000 && adr<0x80200000)
	|| (adr>=0xa0000000 && adr<0xa0200000)
	|| (adr>=0x00000000 && adr<0x00200000)) return ram - (adr&~0x1fffff);

	else if (adr>=0xbfc00000 && adr<0xbfc80000) return rom - 0xbfc00000;
	else if (adr>=0x1f800000 && adr<0x1f801000) return scratchpad - 0x1f800000;
	else return NULL;
}

