#define	SUPPORT_ALPHA
//#define	USE_TCACHE
#define	USE_CCACHE

extern	int verbose;
#define	PRINTF	if (verbose) printf

typedef unsigned long u_long;
typedef unsigned short u_short;
typedef unsigned char u_char;
typedef long fix;
typedef short PIXEL;

typedef struct {
	unsigned char	r,g,b,code;
} COLOR;

typedef struct {
	fix	r,g,b;
} LCOLOR;

typedef struct {
	fix x;
	fix u,v;
	LCOLOR c;
} EDGE;

typedef struct {
	short x,y;
} POINT_F; /* flat point */

typedef struct {
	unsigned char r,g,b,code;
	short x,y;
} POINT_G; /* glow point */

typedef struct {
	short x,y;
	unsigned char u,v; unsigned short pad;
} POINT_FT; /* flat textured point */

typedef struct {
	u_char r,g,b,code;
	short x,y;
	unsigned char u,v; unsigned short pad;
} POINT_GT; /* grow textured point */

#include "prim2.h"

#define	CMASK0	(0x1f)
#define	CMASK1	(0x1f<<5)
#define	CMASK2	(0x1f<<10)
#define	CMASKL0	(0x3f)
#define	CMASKL1	(0x3f<<11)
#define	CMASKL2	(0x3e<<21)

#define	CMASK	(~((1<<15)|(1<<10)|(1<<5)))
#define	CMASKL	(~((1<<31)|(1<<10)|(1<<21)))
#define	COLPACK(c)	((c)&CMASK1)|(((c)>>16)&(CMASK0|CMASK2))
#define	RGB2COL(r,g,b)	(((r)<<23)|((b)<<13)|((g)<<2))
/* 0rrrrrxxxx0bbbbbxxxxx0gggggxxxxx -> 0rrrrrgggggbbbbb */
/* |       |       |       |       |
/* r=1bitȂ */
#define	COLUNPACK(c)	((c)&CMASK1)|(((c)&(CMASK0|CMASK2))<<16)
/* 0rrrrrgggggbbbbb -> 0rrrrr00000bbbbb000000ggggg00000 */

#define	FIX_BIT	10
#define	FIX_1	(1<<FIX_BIT)
#define	I2F(a)	((a)<<FIX_BIT)
#define	F2I(a)	((a)>>FIX_BIT)


static EDGE edge[512*2];
static struct {
	int xmin,ymin,xmax,ymax,tpage;
	int offsetx,offsety;
} gpu = {
	0,0,1023,511,0,0,0
};

static int masktbl[2048];
static PIXEL tmpctbl[256];

extern PIXEL *vram;



void setclip(int xmin,int ymin,int xmax,int ymax)
{
	gpu.xmin = xmin;
	gpu.ymin = ymin;
	gpu.xmax = xmax;
	gpu.ymax = ymax;
}

void settpage(int tpage)
{
	gpu.tpage = tpage;
}

void setoffset(int x,int y)
{
	gpu.offsetx = x;
	gpu.offsety = y;
}

void EdgeF(POINT_F *p0,POINT_F *p1,EDGE *edge)
{
	int dy;
	fix x,dx;

	dy = p1->y - p0->y;
	if (dy==0) return;
	if (dy<0) {
		/* right */
		void *t;
		t = p0; p0=p1; p1=t;
		dy = -dy;
	} else {
		/* left */
		edge++;
	}

	edge+=p0->y*2;
	x = I2F(p0->x);
	dx = I2F(p1->x - p0->x)/dy;

	/* clip */
	if (p1->y<gpu.ymin || p0->y>=gpu.ymax) return;
	if (p0->y<gpu.ymin) {
		int ddy = gpu.ymin - p0->y;
		x += dx*ddy;
		edge +=ddy*2;
		dy-=ddy;
	}
	if (p1->y>=gpu.ymax) {
		dy += gpu.ymax - p1->y;
	}

	do {
		edge->x = x; x+=dx;
		edge+=2;
	} while(--dy);
}

void EdgeFT(POINT_FT *p0,POINT_FT *p1,EDGE *edge)
{
	int dy;
	fix	x,u,v,dx,du,dv;

	dy = p1->y - p0->y;
	if (dy==0) return;
	if (dy<0) {
		/* right */
		void *t;
		t = p0; p0=p1; p1=t;
		dy = -dy;
	} else {
		/* left */
		edge++;
	}

	edge+=p0->y*2;
	x = I2F(p0->x);
	u = I2F(p0->u);
	v = I2F(p0->v);
	dx = I2F(p1->x - p0->x)/dy;
	du = I2F(p1->u - p0->u)/dy;
	dv = I2F(p1->v - p0->v)/dy;

	/* clip */
	if (p1->y<gpu.ymin || p0->y>=gpu.ymax) return;
	if (p0->y<gpu.ymin) {
		int ddy = gpu.ymin - p0->y;
		x += dx*ddy;
		u += du*ddy;
		v += dv*ddy;
		edge +=ddy*2;
		dy-=ddy;
	}
	if (p1->y>=gpu.ymax) {
		dy += gpu.ymax - p1->y;
	}

	do {
		edge->x = x; x+=dx;
		edge->u = u; u+=du;
		edge->v = v; v+=dv;
		edge+=2;
	} while(--dy);
}

#define	ADDCOL(c,dc)	c.r+=dc.r; c.g+=dc.g; c.b+=dc.b
#define	ADDMULCOL(c,dc,m)	c.r+=dc.r*m; c.g+=dc.g*m; c.b+=dc.b*m
#define	CALCDC(dc,c0,c1,d)	dc.r=(c1.r-c0.r)/d;dc.g=(c1.g-c0.g)/d;dc.b=(c1.b-c0.b)/d
#define	RGB2LCOL(c,r0,g0,b0)	c.r=I2F(r0);c.g=I2F(g0);c.b=I2F(b0)
#define	COLOR2PIXEL(c)	((overtbl[c.r>>2]<<10)|(overtbl[c.g>>2]<<5)|(overtbl[c.b>>2]))
#define	LCOL2PIXEL(c)	((overtbl[F2I(c.r)>>2]<<10)|(overtbl[F2I(c.g)>>2]<<5)|(overtbl[F2I(c.b)>>2]))

#define	MULCOL(c,c0) (\
	(overtbl[((((c>>10)&31)*c0.r)/(FIX_1*0x80))]<<10) | \
	(overtbl[((((c>>5)&31)*c0.g)/(FIX_1*0x80))]<<5) | \
	(overtbl[(((c&31))*c0.b)/(FIX_1*0x80)]) )

#define	MULCOL2(c,c0) (\
	(overtbl[((((c>>10)&31)*c0.r)/0x80)]<<10) | \
	(overtbl[((((c>>5)&31)*c0.g)/0x80)]<<5) | \
	(overtbl[(((c&31))*c0.b)/0x80]) )

static unsigned int overtbl[32*2] = {
	 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,
	16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
	31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,
	31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31
};

static PIXEL *shade_ctbl(PIXEL *dist,PIXEL *src,int size,COLOR c0)
{

	PIXEL *ret = dist;
	do {
		int c = *src++;
		*dist++ = MULCOL2(c,c0);
	} while(--size);

	return ret;
}

PIXEL *shade_ctbl2(PIXEL *dist,PIXEL *src,int size,long color)
{
	COLOR c0;
	c0.r = color&255;
	c0.g = (color>>8)&255;
	c0.b = (color>>16)&255;
	return shade_ctbl(dist,src,size,c0);
}

void EdgeG(POINT_G *p0,POINT_G *p1,EDGE *edge)
{
	int	dy;
	fix x,dx;
	LCOLOR c0,c1,dc;

	dy = p1->y - p0->y;
	if (dy==0) return;
	if (dy<0) {
		/* right */
		void *t;
		t = p0; p0=p1; p1=t;
		dy = -dy;
	} else {
		/* left */
		edge++;
	}

	edge+=p0->y*2;
	x = I2F(p0->x);
	dx = I2F(p1->x - p0->x)/dy;
	RGB2LCOL(c0,p0->r,p0->g,p0->b);
	RGB2LCOL(c1,p1->r,p1->g,p1->b);
	CALCDC(dc,c0,c1,dy);

	/* clip */
	if (p1->y<gpu.ymin || p0->y>=gpu.ymax) return;
	if (p0->y<gpu.ymin) {
		int ddy = gpu.ymin - p0->y;
		x += dx*ddy;
		ADDMULCOL(c0,dc,ddy);
		edge +=ddy*2;
		dy-=ddy;
	}
	if (p1->y>=gpu.ymax) {
		dy += gpu.ymax - p1->y;
	}

	do {
		edge->x = x; x+=dx;
		edge->c = c0; ADDCOL(c0,dc);
		edge+=2;
	} while(--dy);
}

void EdgeGT(POINT_GT *p0,POINT_GT *p1,EDGE *edge)
{
	int	y,dy;
	fix	x,u,v,dx,du,dv;
	LCOLOR c0,c1,dc;

	dy = p1->y - p0->y;
	if (dy==0) return;
	if (dy<0) {
		/* right */
		void *t;
		t = p0; p0=p1; p1=t;
		dy = -dy;
	} else {
		/* left */
		edge++;
	}

	edge+=p0->y*2;
	x = I2F(p0->x);
	u = I2F(p0->u);
	v = I2F(p0->v);
	dx = I2F(p1->x - p0->x)/dy;
	du = I2F(p1->u - p0->u)/dy;
	dv = I2F(p1->v - p0->v)/dy;

	RGB2LCOL(c0,p0->r,p0->g,p0->b);
	RGB2LCOL(c1,p1->r,p1->g,p1->b);
	CALCDC(dc,c0,c1,dy);

	/* clip */
	if (p1->y<gpu.ymin || p0->y>=gpu.ymax) return;
	if (p0->y<gpu.ymin) {
		int ddy = gpu.ymin - p0->y;
		x += dx*ddy;
		u += du*ddy;
		v += du*ddy;
		ADDMULCOL(c0,dc,ddy);
		edge +=ddy*2;
		dy-=ddy;
	}
	if (p1->y>=gpu.ymax) {
		dy += gpu.ymax - p1->y;
	}

	do {
		edge->x = x; x+=dx;
		edge->u = u; u+=du;
		edge->v = v; v+=dv;
		edge->c = c0; ADDCOL(c0,dc);
		edge+=2;
	} while(--dy);
}

#define	FRAME_W	1024
#define	FRAME_H	512

#define	ALIGN4(adr)	(((u_long)(adr)&3)==0)

void DrawInit(void)
{
	int i;
		for(i=0;i<2048;i++) {
			int mask = 0;
			if (i&1      ) mask |= 31;
			if (i&(1<<5) ) mask |= 31<<5;
			if (i&(1<<10)) mask |= 31<<10;
			masktbl[i] = mask;
		}

}

void DrawF(EDGE *edge,int ymin,int ymax,unsigned int c)
{
	int dx,dy;
	PIXEL *p,*p0;

	dy = ymax-ymin;
	p0 = vram+ymin*FRAME_W;
	edge += ymin*2;
	c |= (c<<16);
	do {
		dx = F2I(edge[1].x)-F2I(edge[0].x);
		if (dx) {
			p = p0+F2I(edge[0].x);
			if (F2I(edge[0].x)<gpu.xmin) {
				int d = gpu.xmin-F2I(edge[0].x);
				p+=d;
				dx-=d;
			}
			if (F2I(edge[1].x)>gpu.xmax) {
				dx -= F2I(edge[1].x)-gpu.xmax;
			}
			if (dx>0)
#if 1
			do { *p++=c; } while(--dx);
#else
			{
				if (!ALIGN4(p)) *p++=c;
				for(;dx>=2;dx-=2){ *((u_long*)p)++= c;}
				if (dx) *p++=c;
			}
#endif
#if 0

		case T50: /* 50%+50% */
			c &= CMASK;
			do {
				*p = ((*p&CMASK)+c)/2;
				p++;
			} while(--dx);
			break;
		case T100: /* 100%+100% */
			c &= CMASK;
			do {
				u_long cc = (*p&CMASK)+c;
				*p++ = (cc&CMASK) | masktbl[cc>>5];
				/*	0xxxx0xxxx0xxxxx
				+	0xxxx0xxxx0xxxxx
					nxxxxnxxxxnxxxxx
				>>5      nxxxxnxxxxn
				n = 1 Ȃ overflowĂ̂ xxxxx  11111 or
				*/
			} while(--dx);
			break;
		case T100m:	/* 100%-100% */
			c &= CMASK;
			do {
				u_long cc = (*p|~CMASK)-c;
				*p++ = (cc&CMASK) & ~masktbl[cc>>5];
				/*	1xxxx1xxxx1xxxxx
				-	0xxxx0xxxx0xxxxx
					nxxxxnxxxxnxxxxx
				>>5      nxxxxnxxxxn
				n = 0 Ȃ overflowĂ̂ xxxxx  00000 and
				*/
			} while(--dx);
			break;
		case T25: /* 100%+25% */
			c = (c&MKMASK(0x1c))>>2;
			do {
				u_long cc= (*p&CMASK)+c;
				*p++ = (cc&CMASK) | masktbl[cc>>5];
			}
			break;
#endif
		}
		edge+=2;
		p0+=FRAME_W;
	} while(--dy);
}

void DrawG(EDGE *edge,int ymin,int ymax)
{
	int dx,dy;
	PIXEL *p,*p0;
	LCOLOR c0,dc;

	dy = ymax-ymin;
	p0 = vram+ymin*FRAME_W;
	edge += ymin*2;
	do {
		dx = F2I(edge[1].x)-F2I(edge[0].x);
		if (dx) {
			CALCDC(dc,edge[0].c,edge[1].c,dx);
			c0 = edge[0].c;
			p = p0+F2I(edge[0].x);
			if (F2I(edge[0].x)<gpu.xmin) {
				int d = gpu.xmin-F2I(edge[0].x);
				c0.r+=dc.r*d;
				c0.g+=dc.g*d;
				c0.b+=dc.b*d;
				p+=d;
				dx-=d;
			}
			if (F2I(edge[1].x)>gpu.xmax) {
				dx -= F2I(edge[1].x)-gpu.xmax;
			}
			if (dx>0)

			do {
				*p++= LCOL2PIXEL(c0);
				ADDCOL(c0,dc);
			} while(--dx);
		}
		edge+=2;
		p0+=FRAME_W;
	} while(--dy);
}

#define	BITMODE	(3*4)
#define	ABRFLG	(2*16)
#define	BIT4	0
#define	BIT8	(1*4)
#define	BIT15	(2*4)
#define	T50	0|ABRFLG
#define	T100	1|ABRFLG
#define	T100m	2|ABRFLG
#define	T25	3|ABRFLG
#define	NOSHADE	(1*16)

#ifdef USE_CCACHE

extern void *get_clutcache(int clut,int size,int color);

#endif

#ifdef USE_TCACHE

#define	TEXEL4(tex,ctbl,u,v) ctbl[((u_char*)tex)[F2I(u)+F2I(v)*256]]
extern void *get_tcache(int tpage);

#else

#define	TEXEL4(tex,ctbl,u,v) \
	ctbl[(((u_char*)tex)[F2I(u>>1)+F2I(v)*FRAME_W*2]>>((F2I(u)&1)<<2))&15]

#endif

#define	TEXEL8(tex,ctbl,u,v) ctbl[((u_char*)tex)[F2I(u)+F2I(v)*FRAME_W*2]]
#define	TEXEL15(tex,ctbl,u,v)	tex[F2I(u)+F2I(v)*FRAME_W];

#define	ALPHA50(cc,c2) cc = ((cc&CMASK)+(c2&CMASK))/2
#define	ALPHA100(cc,c2)	{ \
	cc = (cc&CMASK) + (c2&CMASK); \
	cc = (cc&CMASK) | masktbl[cc>>5]; }
#define	ALPHA100m(cc,c2) { \
	cc = (c2|~CMASK) - (cc&CMASK); \
	cc = (cc&CMASK) & ~masktbl[cc>>5]; }
#define	ALPHA25(cc,c2) { \
	/* 1100011000111 */ \
	cc = ((cc>>2)&0x18c7) + (c2&CMASK); \
	cc = (cc&CMASK) | masktbl[cc>>5]; }

void DrawFT(EDGE *edge,int ymin,int ymax,int tpage,int clut,COLOR c0code)
{
	int dx,dy,mode;
	PIXEL *p0;
	PIXEL *tex,*ctbl;
	LCOLOR c0;
	if (c0code.r==0x80 && c0code.g==0x80 && c0code.b==0x80)
		c0code.code|=1; /* NOSHADE */

	dy = ymax-ymin;
	p0 = vram+ymin*FRAME_W;
	edge += ymin*2;

#ifdef USE_TCACHE
	if ((tpage&0x180)==0 /*4bit*/) tex = get_tcache(tpage);
	else
#endif
	tex  = vram+(tpage&15)*64+(tpage&16)*(FRAME_W*256/16);
	ctbl = vram+(clut&0x7fff)*16;
	mode = ((tpage>>5)&15)|((c0code.code&3)*16);

	if (!(mode&NOSHADE)) {
		switch(mode&BITMODE) {
#ifdef USE_CCACHE
		case BIT4:
			ctbl = get_clutcache(clut,16,c0code.r|(c0code.g<<8)|(c0code.b<<16));
			break;
		case BIT8:
			ctbl = get_clutcache(clut,256,c0code.r|(c0code.g<<8)|(c0code.b<<16));
			break;
#else
		case BIT4:
			ctbl = shade_ctbl(tmpctbl,ctbl,16,c0code);
			break;
		case BIT8:
			ctbl = shade_ctbl(tmpctbl,ctbl,256,c0code);
			break;
#endif
		case BIT15:
			RGB2LCOL(c0,c0code.r,c0code.g,c0code.b);
			break;
		}
	}

	do {
		dx = F2I(edge[1].x)-F2I(edge[0].x);
		if (dx) {
			fix	u,v,du,dv;
			PIXEL *p;

			u = edge[0].u;
			v = edge[0].v;
			du = (edge[1].u-edge[0].u)/dx;
			dv = (edge[1].v-edge[0].v)/dx;
			p = p0+F2I(edge[0].x);
			if (F2I(edge[0].x)<gpu.xmin) {
				int d = gpu.xmin-F2I(edge[0].x);
				p+=d;
				u+=du*d;
				v+=dv*d;
				dx-=d;
			}
			if (F2I(edge[1].x)>gpu.xmax) {
				dx -= F2I(edge[1].x)-gpu.xmax;
			}
			if (dx>0)
			switch(mode) {
			case BIT4:
			case BIT4|NOSHADE:
				do {
					int c = TEXEL4(tex,ctbl,u,v);
					if (c) *p=c;
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
#ifdef SUPPORT_ALPHA
			case BIT4|T50:
			case BIT4|T50|NOSHADE:
				do {
					int c = TEXEL4(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA50(c,*p);
						*p=c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT4|T100:
			case BIT4|T100|NOSHADE:
				do {
					int c = TEXEL4(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA100(c,*p);
						*p=c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT4|T100m:
			case BIT4|T100m|NOSHADE:
				do {
					int c = TEXEL4(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA100m(c,*p);
						*p=c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT4|T25:
			case BIT4|T25|NOSHADE:
				do {
					int c = TEXEL4(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA25(c,*p);
						*p = c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
#endif
			case BIT8:
			case BIT8|NOSHADE:
				do {
					int c = TEXEL8(tex,ctbl,u,v);
					if (c) *p=c;
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
#ifdef SUPPORT_ALPHA
			case BIT8|T50:
			case BIT8|T50|NOSHADE:
				do {
					int c = TEXEL8(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA50(c,*p);
						*p=c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT8|T100:
			case BIT8|T100|NOSHADE:
				do {
					int c = TEXEL8(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA50(c,*p);
						*p=c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT8|T100m:
			case BIT8|T100m|NOSHADE:
				do {
					int c = TEXEL8(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA100m(c,*p);
						*p=c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT8|T25:
			case BIT8|T25|NOSHADE:
				do {
					int c = TEXEL8(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA25(c,*p);
						*p = c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
#endif
			case BIT15|NOSHADE:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) *p=c;
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
#ifdef SUPPORT_ALPHA
			case BIT15|T50|NOSHADE:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA50(c,*p);
						*p=c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT15|T100|NOSHADE:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA50(c,*p);
						*p=c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT15|T100m|NOSHADE:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA100m(c,*p);
						*p=c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT15|T25|NOSHADE:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						if (c&0x8000) ALPHA25(c,*p);
						*p = c;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
#endif
			case BIT15:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) *p=MULCOL(c,c0);
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
#ifdef SUPPORT_ALPHA
			case BIT15|T50:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA50(c2,*p);
						*p=c2;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT15|T100:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA50(c2,*p);
						*p=c2;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT15|T100m:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA100m(c2,*p);
						*p=c2;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT15|T25:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA25(c2,*p);
						*p = c2;
					}
					p++; u +=du; v+=dv;
				} while(--dx);
				break;
#endif
			}
		}
		edge+=2;
		p0+=FRAME_W;
	} while(--dy);
}

void DrawGT(EDGE *edge,int ymin,int ymax,int tpage,int clut,int code)
{
	int dx,dy,mode;
	fix	u,v,du,dv;
	PIXEL *p,*p0;
	PIXEL *tex,*ctbl;
	LCOLOR c0,dc;

	dy = ymax-ymin;
	p0 = vram+ymin*FRAME_W;
	edge += ymin*2;

#ifdef USE_TCACHE
	if ((tpage&0x180)==0 /*4bit*/) tex = get_tcache(tpage);
	else
#endif
	tex  = vram+(tpage&15)*64+(tpage&16)*(FRAME_W*256/16);
	ctbl = vram+(clut&0x7fff)*16;
	mode = ((tpage>>5)&15)|((code&3)*16);

	do {
		dx = F2I(edge[1].x)-F2I(edge[0].x);
		if (dx) {
			u = edge[0].u;
			v = edge[0].v;
			du = (edge[1].u-edge[1].v)/dx;
			dv = (edge[1].u-edge[1].v)/dy;
			p = p0+F2I(edge[0].x);
			CALCDC(dc,edge[0].c,edge[1].c,dx);
			c0 = edge[0].c;
			if (F2I(edge[0].x)<gpu.xmin) {
				int d = gpu.xmin-F2I(edge[0].x);
				p+=d;
				u+=du*d;
				v+=dv*d;
				c0.r+=dc.r*d;
				c0.g+=dc.g*d;
				c0.b+=dc.b*d;
				dx-=d;
			}
			if (F2I(edge[1].x)>gpu.xmax) {
				dx -= F2I(edge[1].x)-gpu.xmax;
			}
			if (dx>0)
			switch(mode) {
			case BIT4:
				do {
					int c = TEXEL4(tex,ctbl,u,v);
					if (c) *p=MULCOL(c,c0);
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
#ifdef SUPPORT_ALPHA
			case BIT4|T50:
				do {
					int c = TEXEL4(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA50(c2,*p);
						*p=c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT4|T100:
				do {
					int c = TEXEL4(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA50(c2,*p);
						*p=c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT4|T100m:
				do {
					int c = TEXEL4(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA100m(c2,*p);
						*p=c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT4|T25:
				do {
					int c = TEXEL4(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA25(c2,*p);
						*p = c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
#endif
			case BIT8:
				do {
					int c = TEXEL8(tex,ctbl,u,v);
					if (c) *p=MULCOL(c,c0);
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
#ifdef SUPPORT_ALPHA
			case BIT8|T50:
				do {
					int c = TEXEL8(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA50(c2,*p);
						*p=c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT8|T100:
				do {
					int c = TEXEL8(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA50(c2,*p);
						*p=c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT8|T100m:
				do {
					int c = TEXEL8(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA100m(c2,*p);
						*p=c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT8|T25:
				do {
					int c = TEXEL8(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA25(c2,*p);
						*p = c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
#endif
			case BIT15:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) *p=MULCOL(c,c0);
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
#ifdef SUPPORT_ALPHA
			case BIT15|T50:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA50(c2,*p);
						*p=c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT15|T100:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA50(c2,*p);
						*p=c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT15|T100m:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA100m(c2,*p);
						*p=c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
			case BIT15|T25:
				do {
					int c = TEXEL15(tex,ctbl,u,v);
					if (c) {
						int c2 = MULCOL(c,c0);
						if (c&0x8000) ALPHA25(c2,*p);
						*p = c2;
					}
					p++;
					ADDCOL(c0,dc); u +=du; v+=dv;
				} while(--dx);
				break;
#endif
			}
		}
		edge+=2;
		p0+=FRAME_W;
	} while(--dy);
}

#define	ABS(x)	(((x)>0)?(x):-(x))

void DrawLineF(POINT_F *p0,POINT_F *p1,int c)
{
	int dx,dy;
	PIXEL *p;

	dx = p1->x - p0->x;
	dy = p1->y - p0->y;
	if (dx==0 && dy==0) return;
	if (ABS(dx)<=ABS(dy)) {
		int x;
		if (dy<0) {
			void *t;
			t=p0;p0=p1;p1=t;
			dy = -dy;
		}
		dy++;
		p = vram+p0->y*FRAME_W;
		x = I2F(p0->x);
		dx = I2F(p1->x - p0->x +1)/dy;
		do {
			p[F2I(x)] = c; x+=dx;
			p+=FRAME_W;
		} while(--dy);
	} else {
		int y;
		if (dx<0) {
			void *t;
			t=p0;p0=p1;p1=t;
			dx = -dx;
			dy = -dy;
		}
		p = vram+p0->x;
		y = I2F(p0->y);
		dy = I2F(p1->y - p0->y +1)/dx;
		do {
			p[F2I(y)*FRAME_W] = c; y+=dy;
			p++;
		} while(--dx);
	}
}

void DrawLineG(POINT_G *p0,POINT_G *p1)
{
	int dx,dy;
	PIXEL *p;
	LCOLOR	c0,c1,dc;

	dx = p1->x - p0->x;
	dy = p1->y - p0->y;
	if (dx==0 && dy==0) return;
	if (ABS(dx)<=ABS(dy)) {
		int x;
		if (dy<0) {
			void *t;
			t=p0;p0=p1;p1=t;
			dy = -dy;
		}
		dy++;
		RGB2LCOL(c0,p0->r,p0->g,p0->b);
		RGB2LCOL(c1,p1->r,p1->g,p1->b);
		CALCDC(dc,c0,c1,dy);
		p = vram + p0->y*FRAME_W;
		x = I2F(p0->x);
		dx = I2F(p1->x - p0->x +1)/dy;
		do {
			p[F2I(x)] = LCOL2PIXEL(c0);
			x+=dx;
			ADDCOL(c0,dc);
			p+=FRAME_W;
		} while(--dy);

	} else {
		int y;
		if (dx<0) {
			void *t;
			t=p0;p0=p1;p1=t;
			dx = -dx;
		}
		dx++;
		p = vram+p0->x;
		y = I2F(p0->y);
		dy = I2F(p1->y - p0->y +1)/dx;
		RGB2LCOL(c0,p0->r,p0->g,p0->b);
		RGB2LCOL(c1,p1->r,p1->g,p1->b);
		CALCDC(dc,c0,c1,dx);
		do {
			p[F2I(y)*FRAME_W] = LCOL2PIXEL(c0);
			y+=dy;
			ADDCOL(c0,dc);
			p++;
		} while(--dx);
	}
}

void DrawSprite(int x,int y,int w,int h,int u,int v,int tpage,int clut,COLOR c0code)
{
	PIXEL *p0;
	PIXEL *tex,*ctbl;
	int mode;
	LCOLOR c0;

	if (x>gpu.xmax || x+w<gpu.xmin || y>gpu.ymax || y+w<gpu.ymin) return;
	if (x<gpu.xmin) { u+=gpu.xmin-x; x=gpu.xmin; }
	else if (x+w>gpu.xmax) w=gpu.xmax-x;
	if (y<gpu.ymin) { v+=gpu.ymin-x; y=gpu.ymin; }
	else if (y+h>gpu.ymax) h=gpu.ymax-y;

	p0 = vram+y*FRAME_W+x;

#ifdef USE_TCACHE
	if ((tpage&0x180)==0 /*4bit*/) tex = get_tcache(tpage);
	else
#endif
	tex  = vram+(tpage&15)*64+(tpage&16)*(FRAME_W*256/16);
	ctbl = vram+(clut&0x7fff)*16;
	mode = ((tpage>>5)&15)|((c0code.code&3)*16);

	if (!(mode&NOSHADE)) {
		switch(mode&BITMODE) {
#ifdef USE_CCACHE
		case BIT4:
			ctbl = get_clutcache(clut,16,c0code.r|(c0code.g<<8)|(c0code.b<<16));
			break;
		case BIT8:
			ctbl = get_clutcache(clut,256,c0code.r|(c0code.g<<8)|(c0code.b<<16));
			break;
#else
		case BIT4:
			ctbl = shade_ctbl(tmpctbl,ctbl,16,c0code);
			break;
		case BIT8:
			ctbl = shade_ctbl(tmpctbl,ctbl,256,c0code);
			break;
#endif
		case BIT15:
			RGB2LCOL(c0,c0code.r,c0code.g,c0code.b);
			break;
		}
	}

	do {
		PIXEL *p = p0;
		int dx = w;

			switch(mode) {
			case BIT4:
			case BIT4|NOSHADE:
			 {
#ifdef USE_TCACHE
				u_char *p2 = &((u_char*)tex)[u+v*256];
				do {
					int c = ctbl[*p2++];
					if (c) *p = c;
					p++;
				} while(--dx);
#else

#if 0
				int uu = u;
				u_char *p2 = &((u_char*)tex)[v*FRAME_W*2];
				do {
					int c = ctbl[p2[uu/2]>>((uu&1)<<2)&15];
					if (c) *p=c;
					p++; uu++;
				} while(--dx);
#else
				u_char *p2 = &((u_char*)tex)[u/2+v*FRAME_W*2];
				if (u&1) {
					int c = ctbl[(*p2++)>>4];
					if (c) *p = c;
					p++; --dx;
				}
				while(dx>1) {
					int c = ctbl[*p2 & 15];
					if (c) p[0] = c;
					c = ctbl[(*p2++)>>4];
					if (c) p[1] = c;
					p+=2;
					dx-=2;
				}
				if (dx) {
					int c = ctbl[*p2 & 15];
					if (c) *p = c;
				}
#endif

#endif
			 }
				break;
			case BIT8:
			case BIT8|NOSHADE:
			 {
				u_char *p2 = &((u_char*)tex)[u+v*FRAME_W*2];
				do {
					int c = ctbl[*p2++];
					if (c) *p=c;
					p++;
				} while(--dx);
			 }
				break;
			case BIT15|NOSHADE:
			 {
				PIXEL *p2 = &tex[u+v*FRAME_W];
				do {
					int c = *p2++;
					if (c) *p=c;
					p++;
				} while(--dx);
			 }
				break;
			case BIT15:
			 {
				PIXEL *p2 = &tex[u+v*FRAME_W];
				do {
					int c = *p2++;
					if (c) *p=MULCOL(c,c0);
					p++;
				} while(--dx);
			 }
				break;
			}
		v++;
		p0+=FRAME_W;
	} while(--h);
}

void DrawPixel(int x,int y,int c)
{
	if (x>=gpu.xmin && x<=gpu.xmax && y>=gpu.ymin && y<=gpu.ymax)
		vram[y*FRAME_W+x]=c;
}

void DrawBoxNoClip(int x,int y,int w,int h,int c)
{
	PIXEL *p0;
	p0 = &vram[y*FRAME_W+x];
	c = (c<<16)|c;
	do {
		PIXEL *p = p0;
		int dx = w;
		if (!ALIGN4(p)) *p++=c;
		for(;dx>=2;dx-=2){ *((u_long*)p)++= c;}
		if (dx) *p++=c;
		p0+=FRAME_W;
	} while(--h);
}

void DrawBox(int x,int y,int w,int h,int c)
{
	if (x>gpu.xmax || x+w<gpu.xmin || y>gpu.ymax || y+w<gpu.ymin || w==0 || h==00) return;
	if (x<gpu.xmin) x=gpu.xmin;
	else if (x+w>gpu.xmax) w=gpu.xmax-x;
	if (y<gpu.ymin) y=gpu.ymin;
	else if (y+h>gpu.ymax) h=gpu.ymax-y;

	DrawBoxNoClip(x,y,w,h,c);
}


void move(int x0,int y0,int x1,int y1,int w,int h)
{
	PIXEL *from = &vram[y0*FRAME_W + x0];
	PIXEL *to   = &vram[y1*FRAME_W + x1];

	if (from>to) {
		do {
			memcpy(to,from,w*2);
			from += FRAME_W;
			to += FRAME_W;
		} while(--h);
	} else {
		from += FRAME_W*h;
		to += FRAME_W*h;
		do {
			from -= FRAME_W;
			to -= FRAME_W;
			memmove(to,from,w*2);
		} while(--h);
	}
}

#define	MINMAX3() \
	{int xmin,xmax; \
	xmin = xmax = v[0].x; \
	if (xmin>v[1].x) xmin = v[1].x; \
	else if (xmax<v[1].x) xmax = v[1].x; \
	if (xmin>v[2].x) xmin = v[2].x; \
	else if (xmax<v[2].x) xmax = v[2].x; \
	if (xmax==xmin || xmin>=gpu.xmax || xmax<gpu.xmin) return; } \
	ymin = ymax = v[0].y; \
	if (ymin>v[1].y) ymin = v[1].y; \
	else if (ymax<v[1].y) ymax = v[1].y; \
	if (ymin>v[2].y) ymin = v[2].y; \
	else if (ymax<v[2].y) ymax = v[2].y; \
	if (ymax==ymin || ymin>=gpu.ymax || ymax<gpu.ymin) return

#define	MINMAX4() \
	{int xmin,xmax; \
	xmin = xmax = v[0].x; \
	if (xmin>v[1].x) xmin = v[1].x; \
	else if (xmax<v[1].x) xmax = v[1].x; \
	if (xmin>v[2].x) xmin = v[2].x; \
	else if (xmax<v[2].x) xmax = v[2].x; \
	if (xmin>v[3].x) xmin = v[3].x; \
	else if (xmax<v[3].x) xmax = v[3].x; \
	if (xmax==xmin || xmin>=gpu.xmax || xmax<gpu.xmin) return; } \
	ymin = ymax = v[0].y; \
	if (ymin>v[1].y) ymin = v[1].y; \
	else if (ymax<v[1].y) ymax = v[1].y; \
	if (ymin>v[2].y) ymin = v[2].y; \
	else if (ymax<v[2].y) ymax = v[2].y; \
	if (ymin>v[3].y) ymin = v[3].y; \
	else if (ymax<v[3].y) ymax = v[3].y; \
	if (ymax==ymin || ymin>=gpu.ymax || ymax<gpu.ymin) return;


void gpu_polyF3(PolyF3 *prim)
{
	int ymin,ymax;
	POINT_F v[3];
	memcpy(v,prim->v,sizeof(v));

	PRINTF("PolyF3:(%d,%d),(%d,%d),(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y);

	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;

	if (v[0].x*(v[1].y-v[2].y) + v[1].x*(v[2].y-v[0].y) + v[2].x*(v[0].y-v[1].y)<0) {
		POINT_F tmp;
		tmp = v[1]; v[1]=v[2]; v[2]=tmp;
	}

	MINMAX3();
	EdgeF(&v[0],&v[1],edge);
	EdgeF(&v[1],&v[2],edge);
	EdgeF(&v[2],&v[0],edge);
	DrawF(edge,ymin,ymax,COLOR2PIXEL(prim->c));
}

void gpu_polyF4(PolyF4 *prim)
{
	int ymin,ymax;
	POINT_F v[4];
	memcpy(v,prim->v,sizeof(v));

	PRINTF("PolyF4:(%d,%d),(%d,%d),(%d,%d),(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y,v[3].x,v[3].y);

	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	v[3].x += gpu.offsetx;
	v[3].y += gpu.offsety;

	if (v[0].x*(v[1].y-v[2].y) + v[1].x*(v[2].y-v[0].y) + v[2].x*(v[0].y-v[1].y)<0) {
		POINT_F tmp;
		tmp = v[1]; v[1]=v[2]; v[2]=tmp;
	}

	if (v[0].x==v[2].x && v[1].x==v[3].x && v[0].y==v[1].y && v[2].y==v[3].y) {
		int x,y,w,h;
		if (v[0].x<v[1].x) {
			x = v[0].x; w = v[1].x - v[0].x;
		} else {
			x = v[1].x; w = v[0].x - v[1].x;
		}
		if (v[0].y<v[2].y) {
			y = v[0].y; h = v[2].y - v[0].y;
		} else {
			y = v[2].y; h = v[0].y - v[2].y;
		}
		DrawBox(x,y,w,h,COLOR2PIXEL(prim->c));
		return;
	}

	MINMAX4();
	PRINTF("draw %d,%d\n",ymin,ymax);
	EdgeF(&v[0],&v[1],edge);
	EdgeF(&v[1],&v[3],edge);
	EdgeF(&v[3],&v[2],edge);
	EdgeF(&v[2],&v[0],edge);
	DrawF(edge,ymin,ymax,COLOR2PIXEL(prim->c));
}

void gpu_polyG3(PolyG3 *prim)
{
	int ymin,ymax;
	POINT_G v[3];
	memcpy(v,prim->v,sizeof(v));

	PRINTF("PolyG3:(%d,%d),(%d,%d),(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y);

	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	if (v[0].x*(v[1].y-v[2].y) + v[1].x*(v[2].y-v[0].y) + v[2].x*(v[0].y-v[1].y)<0) {
		POINT_G tmp;
		tmp = v[1]; v[1]=v[2]; v[2]=tmp;
	}
	MINMAX3();
	EdgeG(&v[0],&v[1],edge);
	EdgeG(&v[1],&v[2],edge);
	EdgeG(&v[2],&v[0],edge);
	DrawG(edge,ymin,ymax);
}

void gpu_polyG4(PolyG4 *prim)
{
	int ymin,ymax;
	POINT_G v[4];
	memcpy(v,prim->v,sizeof(v));

	PRINTF("PolyG4:(%d,%d),(%d,%d),(%d,%d),(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y,v[3].x,v[3].y);

	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	v[3].x += gpu.offsetx;
	v[3].y += gpu.offsety;
	if (v[0].x*(v[1].y-v[2].y) + v[1].x*(v[2].y-v[0].y) + v[2].x*(v[0].y-v[1].y)<0) {
		POINT_G tmp;
		tmp = v[1]; v[1]=v[2]; v[2]=tmp;
	}
	MINMAX4();
	EdgeG(&v[0],&v[1],edge);
	EdgeG(&v[1],&v[3],edge);
	EdgeG(&v[3],&v[2],edge);
	EdgeG(&v[2],&v[0],edge);
	DrawG(edge,ymin,ymax);
}

void gpu_polyFT3(PolyFT3 *prim)
{
	int ymin,ymax;
	POINT_FT v[3];
	memcpy(v,prim->v,sizeof(v));
	PRINTF("PolyFT3:(%d,%d),(%d,%d),(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y);

	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	if (v[0].x*(v[1].y-v[2].y) + v[1].x*(v[2].y-v[0].y) + v[2].x*(v[0].y-v[1].y)<0) {
		POINT_FT tmp;
		tmp = v[1]; v[1]=v[2]; v[2]=tmp;
	}
	MINMAX3();
	EdgeFT(&v[0],&v[1],edge);
	EdgeFT(&v[1],&v[2],edge);
	EdgeFT(&v[2],&v[0],edge);
	DrawFT(edge,ymin,ymax,prim->v[1].pad,prim->v[0].pad,prim->c);

#if 0
	DrawLineF(&v[0],&v[1],0x7fff);
	DrawLineF(&v[1],&v[2],0x7fff);
	DrawLineF(&v[2],&v[0],0x7fff);
#endif
}

void gpu_polyFT4(PolyFT4 *prim)
{
	int ymin,ymax;
	POINT_FT v[4];
	memcpy(v,prim->v,sizeof(v));
	PRINTF("PolyFT4:(%d,%d),(%d,%d),(%d,%d),(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y,v[3].x,v[3].y);
	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	v[3].x += gpu.offsetx;
	v[3].y += gpu.offsety;
	if (v[0].x*(v[1].y-v[2].y) + v[1].x*(v[2].y-v[0].y) + v[2].x*(v[0].y-v[1].y)<0) {
		POINT_FT tmp;
		tmp = v[1]; v[1]=v[2]; v[2]=tmp;
	}
	MINMAX4();
	EdgeFT(&v[0],&v[1],edge);
	EdgeFT(&v[1],&v[3],edge);
	EdgeFT(&v[3],&v[2],edge);
	EdgeFT(&v[2],&v[0],edge);
	DrawFT(edge,ymin,ymax,prim->v[1].pad,prim->v[0].pad,prim->c);
}

void gpu_polyGT3(PolyGT3 *prim)
{
	int ymin,ymax;
	POINT_GT v[3];
	memcpy(v,prim->v,sizeof(v));
	PRINTF("PolyGT3:(%d,%d),(%d,%d),(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y);
	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	if (v[0].x*(v[1].y-v[2].y) + v[1].x*(v[2].y-v[0].y) + v[2].x*(v[0].y-v[1].y)<0) {
		POINT_GT tmp;
		tmp = v[1]; v[1]=v[2]; v[2]=tmp;
	}
	MINMAX3();
	EdgeGT(&v[0],&v[1],edge);
	EdgeGT(&v[1],&v[2],edge);
	EdgeGT(&v[2],&v[0],edge);
	DrawGT(edge,ymin,ymax,prim->v[1].pad,prim->v[0].pad,prim->v[0].code);
}

void gpu_polyGT4(PolyGT4 *prim)
{
	int ymin,ymax;
	POINT_GT v[4];
	memcpy(v,prim->v,sizeof(v));
	PRINTF("PolyGT4:(%d,%d),(%d,%d),(%d,%d),(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y,v[3].x,v[3].y);

	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	v[3].x += gpu.offsetx;
	v[3].y += gpu.offsety;
	if (v[0].x*(v[1].y-v[2].y) + v[1].x*(v[2].y-v[0].y) + v[2].x*(v[0].y-v[1].y)<0) {
		POINT_GT tmp;
		tmp = v[1]; v[1]=v[2]; v[2]=tmp;
	}
	MINMAX4();
	EdgeGT(&v[0],&v[1],edge);
	EdgeGT(&v[1],&v[3],edge);
	EdgeGT(&v[3],&v[2],edge);
	EdgeGT(&v[2],&v[0],edge);
	DrawGT(edge,ymin,ymax,prim->v[1].pad,prim->v[0].pad,prim->v[0].code);
}

void gpu_lineF2(LineF2 *prim)
{
	POINT_F v[2];
	memcpy(v,prim->v,sizeof(v));
	PRINTF("LineF2:(%d,%d)-(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y);

	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	DrawLineF(&v[0],&v[1],COLOR2PIXEL(prim->c));
}

void gpu_lineF3(LineF3 *prim)
{
	POINT_F v[3];
	PRINTF("LineF3:(%d,%d)-(%d,%d)-(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y);
	memcpy(v,prim->v,sizeof(v));
	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	DrawLineF(&v[0],&v[1],COLOR2PIXEL(prim->c));
	DrawLineF(&v[1],&v[2],COLOR2PIXEL(prim->c));
}

void gpu_lineF4(LineF4 *prim)
{
	POINT_F v[4];
	memcpy(v,prim->v,sizeof(v));

	PRINTF("LineF4:(%d,%d)-(%d,%d)-(%d,%d)-(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y,v[3].x,v[3].y);

	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	v[3].x += gpu.offsetx;
	v[3].y += gpu.offsety;
	DrawLineF(&v[0],&v[1],COLOR2PIXEL(prim->c));
	DrawLineF(&v[1],&v[2],COLOR2PIXEL(prim->c));
	DrawLineF(&v[2],&v[3],COLOR2PIXEL(prim->c));
}

void gpu_lineG2(LineG2 *prim)
{
	POINT_G v[2];
	memcpy(v,prim->v,sizeof(v));
	PRINTF("LineG2:(%d,%d)-(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y);
	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	DrawLineG(&v[0],&v[1]);
}

void gpu_lineG3(LineG3 *prim)
{
	POINT_G v[3];
	memcpy(v,prim->v,sizeof(v));
	PRINTF("LineG3:(%d,%d)-(%d,%d)-(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y);
	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	DrawLineG(&v[0],&v[1]);
	DrawLineG(&v[1],&v[2]);
}

void gpu_lineG4(LineG4 *prim)
{
	POINT_G v[4];
	memcpy(v,prim->v,sizeof(v));
	PRINTF("LineG4:(%d,%d)-(%d,%d)-(%d,%d)-(%d,%d)\n",v[0].x,v[0].y,v[1].x,v[1].y,v[2].x,v[2].y,v[3].x,v[3].y);
	v[0].x += gpu.offsetx;
	v[0].y += gpu.offsety;
	v[1].x += gpu.offsetx;
	v[1].y += gpu.offsety;
	v[2].x += gpu.offsetx;
	v[2].y += gpu.offsety;
	v[3].x += gpu.offsetx;
	v[3].y += gpu.offsety;
	DrawLineG(&v[0],&v[1]);
	DrawLineG(&v[1],&v[2]);
	DrawLineG(&v[2],&v[3]);
}

void gpu_fill(Fill *prim)
{
	PRINTF("Fill:(%d,%d) (%dx%d)\n",prim->v[0].x,prim->v[0].y,prim->w,prim->h);
	DrawBoxNoClip(prim->v[0].x,prim->v[0].y,prim->w,prim->h,COLOR2PIXEL(prim->c));
}

void gpu_tile(Tile *prim)
{
	PRINTF("Tile:(%d,%d) (%dx%d)\n",prim->v[0].x,prim->v[0].y,prim->w,prim->h);
	DrawBox(prim->v[0].x+gpu.offsetx,prim->v[0].y+gpu.offsety,prim->w,prim->h,COLOR2PIXEL(prim->c));
}

void gpu_tile8(Tile8 *prim)
{
	PRINTF("Tile8:(%d,%d)\n");
	DrawBox(prim->v[0].x+gpu.offsetx,prim->v[0].y+gpu.offsety,8,8,COLOR2PIXEL(prim->c));
}

void gpu_tile16(Tile16 *prim)
{
	PRINTF("Tile16:(%d,%d)\n");
	DrawBox(prim->v[0].x+gpu.offsetx,prim->v[0].y+gpu.offsety,16,16,COLOR2PIXEL(prim->c));
}

void gpu_tile1(Tile1 *prim)
{
	PRINTF("Tile1:(%d,%d)\n");
	DrawPixel(prim->v[0].x+gpu.offsetx,prim->v[0].y+gpu.offsety,COLOR2PIXEL(prim->c));
}

void gpu_sprt(Sprt *prim)
{
	PRINTF("Sprt:(%d,%d) (%dx%d)\n",prim->v[0].x,prim->v[0].y,prim->w,prim->h);
	DrawSprite(prim->v[0].x+gpu.offsetx,prim->v[0].y+gpu.offsety,prim->w,prim->h,
	prim->v[0].u,prim->v[0].v,gpu.tpage,prim->v[0].pad,prim->c);
}

void gpu_sprt8(Sprt *prim)
{
	PRINTF("Sprt8:(%d,%d)\n");
	DrawSprite(prim->v[0].x+gpu.offsetx,prim->v[0].y+gpu.offsety,8,8,
	prim->v[0].u,prim->v[0].v,gpu.tpage,prim->v[0].pad,prim->c);
}

void gpu_sprt16(Sprt *prim)
{
	PRINTF("Sprt16:(%d,%d)\n");
	DrawSprite(prim->v[0].x+gpu.offsetx,prim->v[0].y+gpu.offsety,16,16,
	prim->v[0].u,prim->v[0].v,gpu.tpage,prim->v[0].pad,prim->c);
}
